// ################################################################################
//
//  Panel 48x128
//
//  Released:  30/11/2021
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    Micro:  ESP32

    This code controls a 48x16 RGB panel, using a controller, which has a rotary
    encoder, an SD memory card and OLED display.
    Functions provided are as follows:
    * Menu system
      * selects animation
      * controls brightness
      * controls speed
    * Auto runs An0000.txt if it exists
*/
// Declare Libraries
#include <Arduino.h>          // needed when using an ESP32 micro
#include <HardwareSerial.h>   // serial library
#include <Wire.h>             // I2C library
#include <FastLED.h>          // Neopixel library
#include "SSD1306Wire.h"      // OLED display library
#include "FS.h"               // file system library
#include "SD.h"               // SD card library
#include "SPI.h"              // SPI library
#include <stdlib.h>

// IMPORTANT - the following values configure the micro for the LED panels
// set these numbers correctly to ensure Gfx functions work as expected
#define PixH 1 * 16               // number of vertical pixels (v panel x 16) 
#define PixW 3 * 16               // number of horizontakl pixels (h panels x 16)
#define NumLEDs (PixH * PixW) + 1 // total number of LEDs
#define PixWxPixH PixW * PixH     // max array element

// Configuration
String Release = "Panel48x128 (R3)";
String Date = "30/11/2021";
#define BUILD true // == true if display and rotary encoder are connected

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 21;         // I2C SDA GPIO pin
const int SCL_PIN = 22;         // I2C SCL GPIO pin

// Initialize the OLED display for address 0x3c
SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SCL_PIN);

// Define a task for core 0
TaskHandle_t Core0;

// Define constants
#define ACK 6           // link acknowledge charcater
#define CR 13           // carriage return charcater
#define Debounce 5      // debounce time in milliseconds
#define DEL 127         // delimeter character ASCII (127)
#define LED_Pin 4       // GPIO 4 connected to first LED
#define LEDSwMax 255    // max LED brightness for switch operations
#define LF 10           // line feed charcater
#define sw0Pin 13       // GPIO13 is connected to SW0 on rotary encoder
#define swAPin 27       // GPIO27 is connected to A on rotary encoder
#define swBPin 26       // GPIO27 is connected to B on rotary encoder
#define TESTmode false  // true during assembly, false for normal use
#define timeout 500     // rotary encoder symbol timeout
#define Xmax PixW-1     // maximum value of X in co-ordinate system
#define Xmin 0          // minimum value of X in co-ordinate system
#define Ymax PixH-1     // maximum value of Y in co-ordinate system
#define Ymin 0          // minimum value of Y in co-ordinate system
 
// create a frame FastLED instance and define the total number of LEDs
CRGB LED[NumLEDs];
CRGB Forgnd[NumLEDs];
CRGB Bckgnd[NumLEDs];

// creat FastLED data elements for the Sprite16 images
// note that Matrix editor assumes LEDs start at 1 not zero, hence 257 LEds
CRGB Sprite0[257];  // sprite 0
CRGB Sprite1[257];  // sprite 1
CRGB Sprite2[257];  // sprite 2
CRGB Sprite3[257];  // sprite 3
CRGB Sprite4[257];  // sprite 4
CRGB Sprite5[257];  // sprite 5
CRGB Sprite6[257];  // sprite 6
CRGB Sprite7[257];  // sprite 7

// create FastLED data element for Gfx functions. These are never shown to a GPIO pin
CRGB BakRGB[1];     // background pixel colour
CRGB Block[128];    // temp store for pixel manipulation
CRGB ColRGB[1];     // foreground pixel colour
CRGB MemRGB[1];     // a memory for storing a colour temporarily

// Declare SD reader variables
uint8_t cardType;   // numeric ref to card type
bool ErrSD;         // general error flag
bool SDin = false;  // == true if SD card is attached

// Declare and initialise global variables
int16_t ACK_Cnt;    // counter used in app link connection
bool AnimBuild;     // true if the animation engine is set to run
int16_t AnimDel;    // if > 0 then skip frame cycles
int16_t AnimDp;     // animation data pointer for serial comms
int16_t AnimErr;    // if > 0 it references an animation error
int32_t AnimFrame;  // animation frame loaded, -1 not loaded
bool AnimIf;        // do next line, default = true
int16_t AnimInc;    // animation engines pointer increment
int16_t AnimLast;   // previous AnimMode
int16_t AnimLoop;   // pointer to Loop return line in animations
int16_t AnimMode;   // a pointer which determines which frame build process is used
bool AnimPacket;    // if == true then decode serial data as Anim packets
bool AnimPAUSE;     // if == true then the current animation is paused
bool AnimPLAY;      // == true if an animation has been set to play
int16_t AnimPnt;    // animation engines line pointer
bool AnimSTEP;      // animation engine single STEP invokes a pause
bool AnimSTOP;      // set == true to stop current animation
unsigned long AnimWait; // time delay in millis()
bool AnyErr;        // tempr error test
int AnyInt;         // temp integer value
String AnyStr = ""; // temp string variable
uint8_t BAK_Blu;    // current background blue colour
uint8_t BAK_Grn;    // current background green colour
uint8_t BAK_Red;    // current background red colour
int32_t BckGndFrame;// background frame loaded, -1 not loaded
bool BootAnim;      // set = true if animation is running from boot
bool Border;        // if true then display has a border
uint8_t Bright;     // FastLED brightness level
bool BrightAuto;    // if == true automatically increase brightness to max levels
uint8_t BrightCnt;  // counter used in auto max
uint8_t BrightDef;  // default FastLED brightness level
uint8_t BrightLast; // previous brightness value
uint8_t Brightness; // OLED display brightness
uint8_t BrightNew;  // Latest FastLED brightness
char cmdMode;       // command mode
String cmdRx;       // data received from serial port, excluding sopecial characters
int16_t cmdSgn;     // cmdVal sign, 0 or -1
char cmdType;       // command mode type
int16_t cmdVal;     // value associated with a cmdType
uint8_t Col11;      // max colour
uint8_t Col34;      // 3/4 max colour
uint8_t Col12;      // 1/2 maxcolour
uint8_t Col14;      // 1/4 maxcolour
byte Col_Pnt;       // colour pointer
int16_t DELcnt;     // counts the delimeters in the Rx data stream
bool DispBack;      // flag used to draw a back arrow on the display
bool DispChng;      // flag invokes initialisation code
bool DispClr;       // set == true to clear the memory contents
int16_t DispCnt;    // counter used in display updating
byte DispCX,DispCY; // display cursor coordinates
int16_t DispDel;    // >0 display delay counter, stops updates over I2C
bool DispDisp;      // set = true to immediately display memory contents
bool DispLock;      // set == true to lock the display
byte DispMenu;      // group type, 0 = sensors, 1 = tasks, etc
int16_t DispMode;   // the type of display for a given mode
bool DispNOP;       // set == true to stop display I2C updates
int16_t DispNow;    // if > 0 display special modes
bool DispOnce;      // used to set DispTask initial conditions
int16_t DispPnt;    // points at the current display mode
int16_t DispSub;    // special display sub-task pointer
bool DispSW0;       // the state of SW0 when the task was initiated
int16_t DispTask;   // special display task pointer
bool DispYN;        // display yes/no flag
bool ErrFlg;        // general error flag used in file reading
bool ESC0;          // SW0 ESC exit current Main Task flag
String Ext;         // extension string. eg. '.txt'
String Folder;      // folder name useed in filepath read/write
int16_t ForAdd[10]; // For... Next loop address pointers
int16_t ForCnt[10]; // For... Next loop counters
int32_t ForGndFrame;  // foreground frame loaded, -1 not loaded
int16_t ForPnt;     // For... Next loop stack pointer
bool FrameBuild;    // if == true then run a frame building task
int16_t FrameCnt;   // Main Task general counter/timer
int16_t FrameDel;   // Main task delay counter, if > 0 holds off tasks
String FrameMode;   // text describing the frame mode used by the editor
int16_t FrameNext;  // next FrameTask to be performed after a delay
int16_t FrameNum;   // current FrameTask to be performed after a delay
unsigned long FramePeriod;    // frame time period in ms, default = 40 (25Hz)
unsigned long FramePeriodDef; // frame time period in ms, default = 40 (25Hz)
unsigned long FramePeriodLast;// previous frame time period in ms, default = 40 (25Hz)
bool FrameRESET;    // if == true then stop build tasks running
int8_t FrameRnd;    // any random number 0 - 255
int16_t FrameSubTask;       // general subtask pointer
int16_t FrameTask;  // task pointer for frame construction
unsigned long FrameTimer;   // frame timer in ms
String FrameTitle;  // text describing the frame title used by the editor
byte GB[8];         // temp buffer for Gfx characters
byte GC[8];         // buffer for Gfx characters
uint8_t Gfx_Char;   // pointer to a Gfx character
int16_t GfxMax;     // points to last element of current face, top left corner
int16_t GfxP0;      // points to first element of current face, bottom right corner
byte GH,GW;         // Gfx character height and width
int16_t Gosub[10];  // Gosub return stack
int16_t GosubPnt;   // Gosub stack pointer
String Hex;         // any hex string
int16_t IcMax;      // if > 0 the use IcMax/10 as current limit
bool IcOnce;        // set == true after each show for current measurement
char keyChar;       // any keyboard character
int16_t keyVal;     // any keyboard value
bool LED_Blink;     // flag toggled to create blinking effects
uint8_t LED_Char;   // general char value used in LED tasks
uint8_t LED_Col;    // general colour pointer used in LED tasks
int16_t LED_Cnt;    // general counter used in LED tasks
int16_t LED_Del;    // LED Task delay, adds pauses into the processes
uint8_t LED_End;    // general end counter used in LED tasks
//int8_t LED_Face;    // cube face 0 - 4
bool LED_Flag;      // any true/false flag
int16_t LED_Inc;    // general increment used in LED tasks
int16_t LED_Last;   // previous LED task pointer
uint8_t LED_Max;    // Max brightness level used by tasks
uint8_t LED_Min;    // general min number used by tasks
bool LEDnoshow;     // temporary flag used to disable frame output
bool LED_Nop;       // == true to block LED tasks, default = false
int16_t LED_Num;    // general number stored used in LED tasks
bool LED_ON;        // built-in LED ON flag
int16_t LED_Pnt;    // pointer to LED to be lit
int16_t LED_PXY;    // LED array pointer
uint8_t LED_Rnd;    // 8-bit random number
bool LEDshow;       // LED show flag, triggers FastLED.show()
int16_t LED_SubTask;// LED sub-task pointer
bool LED_SwDwn;     // == true if a button switch has been pressed
int16_t LED_Task;   // LED task pointer
String LED_Text;    // any string used in LED tasks
byte LED_Val;       // values used in LED tasks
uint8_t LED_X;      // any X face co-ordinates
uint8_t LED_Y;      // any Y face co-ordinates
int LevelCnt;       // counts cumulative levels of all LEDs in output chain
int LevelOld;       // previous cumulative levels of all LEDs in output chain
unsigned long loopDelay;  // RTOS task timer
uint8_t Mem_Blu;    // memory of current blue colour
uint8_t Mem_Grn;    // memory of current green colour
uint8_t Mem_Red;    // memory of current red colour
uint8_t MenuLevel;  // pointer to menu level in the hierarchy, 0 = top, - N
uint8_t MenuPnt;    // pointer to the current tier level, 0 - N
uint8_t MenuT0;     // menu choice range LL
uint8_t MenuT1;     // menu choice range LL
uint8_t MenuTL;     // menu tier list 4th item pointer
uint8_t MenuTS;     // menu tier list start pointer
String Name;        // filename useed in filepath read/write, minus .txt extension
unsigned long next5ms;    //  5ms loop timer used for reading the rotary encoder
unsigned long next10ms;   //  10ms loop timer
unsigned long next40ms;   //  40ms loop timer
unsigned long PauseTimer; // Anim Pause command wait timer
uint8_t Pen_Blu;    // current pen blue colour
uint8_t Pen_Grn;    // current pen green colour
uint8_t Pen_Red;    // current pen red colour
bool PrintEn;       // if == false turn off verbose file system messaging
bool REBOOT;        // set true when reboot is in progress to stop all tasks
bool RepON;         // report ON/OFF flag, default = true
byte ResetCnt;      // timer used to invoke soft RESET
uint8_t RndCol;     // previous random colour
int16_t RotCnt = 0; // symbol duration timer
int16_t RotEncCnt;  // rotary encoder counter
int16_t RotState = 0;   // symbol state
bool RotUpd;        // == true when a count update has occured
int16_t SchCnt;     // animation scheduler array counter, default 0
int16_t SchPnt;     // animation scheduler array pointer, default -1
int16_t ScrollTask; // == 0 to cause scroll initialisation
int16_t SpriteNum;  // sprite file number, SpNNNN.txt
int16_t SpritePnt;  // sprite array pointer 0 - 7
int16_t sw0Cnt;     // button switch counter
int16_t sw0DwnTime; // button pressed down time
bool sw0LastState;  // previous state of button switch, HIGH/LOW
bool SW0_Nop;       // == true to block SW0 read tasks, default = false
bool sw0State;      // state of read button switch pin
int16_t sw0Timer;   // timer used to detemine button sequences
bool swAState;      // state of swA input
bool swBState;      // state of swB input
unsigned long t0,t1;  // temp timers for function tests
int16_t Task10ms;   // 10ms loop timer tasks
int16_t Task40ms;   // 40ms loop timer tasks
bool TEST;          // normally false, set = true for assembly test mode
String Text;        // any text to be written to the display
int16_t Text_SY;    // Y position of scrolling text
int16_t Text_X;     // text X co-ordinate
int16_t Text_Y;     // text Y co-ordinate
uint8_t Txt_Blu;    // current text blue colour
uint8_t Txt_Char;   // character value used in text scrolling
int16_t Txt_Cnt;    // counter used in text scrolling
uint8_t Txt_Grn;    // current text green colour
int16_t Txt_Num;    // number of characters in scrolling text
int16_t Txt_Pnt;    // pointer for scrolling text
uint8_t Txt_Red;    // current text red colour
int16_t UsbActive;  // > 0 if serial data is received
bool UsbState;      // == true if in UsbActive mode
bool Verbose;       // if == true send all text messages
bool WaitUp;        // true if task is to wait for button release
String Word;        // any word extracted from Text to be written to the display
int16_t WordNo;     // pointer to the ~ char in a text string
String WordRht;     // any remaining word extracted from Text to be written to the display
int16_t WordSp;     // pointer to the spaces in a text string between words

// create animation arrays
const int AnimDepth = 1000;   // max number of array elements
uint16_t AnimMax = AnimDepth; // number of elements store in the array
const int16_t SchDepth = 100; // depth of scheduler references

uint16_t AnCmd[AnimDepth];    // animation commands
int32_t AnD0[AnimDepth];      // animation data element 0
int32_t AnD1[AnimDepth];      // animation data element 1
int32_t AnD2[AnimDepth];      // animation data element 2
String  AnStr[AnimDepth];     // string elements

int16_t AnimSch[SchDepth];    // animation scheduler file numbers, ie. AnXXXX.text

int16_t SpriteDiv[8] = {1,1,1,1,1,1,1,1}; // divider values
int16_t SpriteFN[8] = {-1,-1,-1,-1,-1,-1,-1,-1};  // loaded file references
int16_t SpriteInc[8] = {0,0,0,0,0,0,0,0}; // increment to target values
bool SpriteON[8] = {false,false,false,false,false,false,false,false};
int16_t SpriteSX[8] = {0,0,0,0,0,0,0,0};  // start X values
int16_t SpriteSY[8] = {0,0,0,0,0,0,0,0};  // start Y values
int16_t SpriteTX[8] = {0,0,0,0,0,0,0,0};  // target X values
int16_t SpriteTY[8] = {0,0,0,0,0,0,0,0};  // target Y values
int16_t SpriteX[8] = {0,0,0,0,0,0,0,0};   // X co-ordinates
int16_t SpriteY[8] = {0,0,0,0,0,0,0,0};   // Y co-ordinates

int16_t Var[26];              // variables 'A' - 'Z'

// --------------------------------------------------------------------------------

void setup() {
  // Setup code, runs once:
  pinMode(sw0Pin,INPUT_PULLUP); // attach internal pull-up to rotary encoder pin SW0
  pinMode(swAPin,INPUT_PULLUP); // attach internal pull-up to rotary encoder pin A
  pinMode(swBPin,INPUT_PULLUP); // attach internal pull-up to rotary encoder pin B

  setDefaults();
  
  Serial.begin(500000); // baud rate for serial monitor USB Rx/Tx comms

  // initialise the FastLED component
  FastLED.addLeds<WS2812B, LED_Pin, GRB>(LED, NumLEDs);
  FastLED.setBrightness(BrightDef);
 
  RunPOST();  // do some basic checks

  // set loop timers for multi-tasking events BEFORE activating Core 0
  synchLoopTimers();

  // define the Core 0 task parameters
  // Core 0 is normally used for WiFi, but here we use it to read sensors
  // leaving Core 1 to create the LED patterns and clock them out
  // Note, this code starts and runs immediately, before loop() does
  xTaskCreatePinnedToCore(
    loop0,       // Task function.
    "Core0",     // name of task.
    10000,       // Stack size of task
    NULL,        // parameter of the task
    1,           // priority of the task
    &Core0,      // Task handle to keep track of created task
    0);          // pin task to core 0
}

// ---------------------------------------------------------------------------------

  // the controller will react to a soft RESET event when SW0 is held down
  // this function effectively restarts the code as if from power-up
void(* resetFunc) (void) = 0; //declare reset function at address 0

// ---------------------------------------------------------------------------------

void loop() {
  // main code which runs repeatedly as multi-tasking functions
  // this code is running in Core 1 and focuses on LED graphics

  //###############################################################################
  //
  //  Frame tasks 20+ms (<=50Hz) variable
  //
  //###############################################################################
  // These tasks are run at a variable rate, so that LED effects can be speeded up
  // and slowed down in a smooth manner by changing the variable FramePeriod
  // when the frame timer triggers it 'shows' the LEDs, if they are set to be shown
  // then it sets about building the next frame during the following period
  //###############################################################################
  // Build
  //###############################################################################
  if (AnimBuild) {
    // The animation engine has been invoked, which takes priority over frame building tasks
    AnimEngine();
    // at the end of building animation frames we add sprites if ON
    if (!AnimBuild) {
      // animation build has concluded
      IcOnce = true;  // set flag to perform current measurement
      if (AnimMode == 1) {
        // apply active sprites to output frame
        for (int zSP = 0;zSP < 8;zSP++) {
          if (SpriteON[zSP]) {
            // sprite is ON so load it at SpriteX[zSP],SpriteY[zSP]
            Gfx_SpriteLoad(zSP);
          }
        }
      }
    }
  } else if (FrameBuild) {
    // whilst FrameBuild == true we construct the new frame, then clear FrameBuild
    doFrameTasks();
//    t1 = micros();
  }
  //###############################################################################
  // Frame IcMax current count
  //###############################################################################
  else if (AnimPLAY && IcOnce) {
    // only do current measurement if IcOnce == true
    // on a 48x16  panel this takes between 172us - 232us for all off to all on
    // on a 48x128 panel this takes between 501us - 738us for all off to all on
//    t0 = micros();
    IcOnce = false; // bloack this check until the next show event
    LevelCnt = 0;   // clear the LED level counter
    if (IcMax > 0) {
      // a current limit has been set so we scan the output frame for ON pixels
      MemRGB[0].setRGB(0,0,0);  // set the black reference OFF colour
      uint8_t zBN = BrightNew;  // note current brightness level
      for (int zL = 1; zL < NumLEDs;zL++) {
        if (LED[zL] != MemRGB[0]) {
          // LED value is not zero, so count its colour contribution
          LevelCnt+= LED[zL].r; LevelCnt+= LED[zL].g; LevelCnt+= LED[zL].b;
        }
      }
      if (AnimPLAY && (LevelOld != LevelCnt)) {
        // current level is different to previous value
        float zIb = (0.02 * BrightNew * LevelCnt)/65025.0;  // current at current brightness
        float zIc = (0.02 * LevelCnt)/255.0;                // current @ Bright == 255
        float zIcMax = 0.1 * IcMax;
        if (zIc > zIcMax) {
          Bright = (zIcMax * 255)/zIc;  // pull down brightness to not exceed IcMax
          if (Bright < BrightNew) {BrightNew = Bright; FastLED.setBrightness(BrightNew); BrightCnt = 0;}
        }
        if (Verbose) {Serial.println(String(LevelCnt) + "  " + String(zIc) + " Amps  " + String(BrightNew) + "  " + String((0.02 * BrightNew * LevelCnt)/65025) + " Amps");}
      }
      LevelOld = LevelCnt;
      if (BrightNew > 0) {  // avoid divide by zero
        int16_t zLED0 = 1 + (255/BrightNew);
        if (BrightNew < zBN) {LED[0].setRGB(zLED0,0,0);} // turn micros LED red
        else {LED[0].setRGB(0,zLED0,0);} // turn micros LED green
      }
    }
//    t1 = micros(); Serial.println(t1 - t0);
  }
  //###############################################################################
  // Frame timer
  //###############################################################################
  else if ((millis() - FrameTimer) >= FramePeriod) {
    // we have reached the end of the FrameTimer period
    // if FrameBuild tasks resulted in graphics then LEDshow will be = true
    FrameTimer = millis();  // set the timer for the next period
//    Serial.println(t1 - t0);
    if (LEDnoshow) {
        // flag is set when switching between frame building and animation
        // cancel pending frame show
        LEDshow = false; LEDnoshow = false;
    }
    if (AnimDel > 0) {AnimDel--;} // AnimDel > 0 blocks 'show' and 'build' cycles
    else {
      //###############################################################################
      // Show
      //###############################################################################
      if (LEDshow) {
        // update RGB LEDs if LEDshow == true
        // this consumes Core 1 processing until finished
        LEDshow = false;  // clear the flag early in case it is set whilst showing
        if (AnimSTEP && !AnimPAUSE) {AnimSTEP = false; AnimPAUSE = true;}
//        t0 = micros();
        FastLED.show(); // t1 - t0 = 188828 for 6145 LEDs, approx 5.3Hz
//        t1 = micros(); Serial.println(t1 - t0);
      }
      // after updating the LEDs we need to build the next output
      // the route taken will depend on AnimMode, if -1 do nothing
      if (AnimMode == 0) {
        // default mode is frame building from tasks
        FrameBuild = true;  // initiate the next frame build process of tasks
      } else if (AnimMode == 1) {
        // AnimMode is set to play an animation script loaded into memory
        // continue playing the script unless a PAUSE is applied
        if (AnimPLAY && !AnimPAUSE) {AnimBuild = true;}
      }
    }
//    t0 = micros();
  }
  // if COMs is suddenly connected we stop all animation engines
  if (AnimSTOP) {
    Serial.println("AnimSTOP");
    ResetAnimEngines(); AnimSTOP = false;
  }


  //###############################################################################
  //
  //  40 ms Tasks (25Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 40ms period
  if (Task40ms > 0) {
    switch(Task40ms) {
      case  1: Display_40ms(); break;  // update display if necessary
    } Task40ms--;
  } else if ((millis() - next40ms) >= 40) {
    next40ms = millis(); Task40ms = 1;
  }
}

// ---------------------------------------------------------------------------------

void loop0(void * pvParameters) {
  // this function runs on Core 0, normally used for ESP32 WiFi
  // *** WARNING *** do not use for time critical code like I2C, FastLED, etc
  // the following message, sent once, confirms core assignment
  Serial.print("Core0 task running on core "); Serial.println(xPortGetCoreID());

  for(;;){
    // infinite loop, just like loop() but running in Core 0
    // you must never use 'break' or 'return' statement here or it will crash
  
    // whilst serial data is arriving put back loop task release
    while (Serial.available() > 0) {
      // read serial port on until it is empty
      readSerial(); // check serial port for incoming data from Matrix editor
    }
    
    //###############################################################################
    //
    //  5 ms rotary encoder task (200Hz)
    //
    //###############################################################################
    // this code generates a count value for the rotary encoder, RotEncCnt
    // when the count changes an update flag is set, RotUp = true
    if (!UsbState && BUILD && ((millis() - next5ms) >= Debounce)) {
      // every 5ms, set by Debounce, we check the state of the A and B pins
      next5ms = millis();
      swAState = digitalRead(swAPin); swBState = digitalRead(swBPin);
      switch (RotState) {
        case -1:  // error state due to symbol timeout
          // wait for both pins to return to HIGH state
          if (swAState && swBState) {RotState = 0; RotCnt= 0;}
          break;
        case 0: // waiting for a transition to occur
                if ((!swAState) && (!swBState)) {RotState = -1;}  // both pins gone LOW, timing ERROR
          else if (!swAState) {RotState = 1;}      // pin A has gone LOW, turning clockwise
          else if (!swBState) {RotState = 4;} // pin B has gone LOW, turning anti-clockwise
          break;
        case 1: // clockwise rotation
          // A == LOW, B == HIGH waiting for the next transition
          if (!swBState) {RotState = 2;}  // expected transition of B == LOW
          else if (swAState) {RotState = 0; RotCnt= 0;} // unexpected bounce on pin A
          break;
        case 2:  // clockwise rotation
          // A == LOW, B == LOW waiting for the next transition
          if (swAState) {RotState = 3;}       // expected transition of A == HIGH
          else if (swBState) {RotState = 1;}  // unexpected bounce on pin B
          break;
        case 3: // clockwise rotation
          // A == HIGH, B == LOW waiting for next transition
          if (swBState) {RotState = 0; RotCnt = 0; RotEncCnt++; RotUpd = true;} // expected transition of B == HIGH
          else if (!swAState) {RotState = 2;} // unexpected bounce on pin A
          break;
        case 4: // anti-clockwise rotation
          // A == HIGH, B == LOW waiting for the next transition
          if (!swAState) {RotState = 5;}  // expected transition of B == LOW
          else if (swBState) {RotState = 0; RotCnt= 0;} // unexpected bounce on pin B
          break;
        case 5: // anti-clockwise rotation
          // A == LOW, B == LOW waiting for the next transition
          if (swBState) {RotState = 6;}       // expected transition of B == HIGH
          else if (swAState) {RotState = 4;}  // unexpected bounce on pin A
          break;
        case 6: // anti-clockwise rotation
          // A == LOW, B == HIGH waiting for next transition
          if (swAState) {RotState = 0; RotCnt = 0; RotEncCnt--; RotUpd = true;} // expected transition of B == HIGH
          else if (!swBState) {RotState = 5;} // unexpected bounce on pin B
          break;
      }
      // check for timeout conditions
      if (RotState > 0) {RotCnt++; if (RotCnt > timeout) {RotState = -1;}}
  //    Serial.println(RotState);
    }

    
    //###############################################################################
    //
    //  10 ms non-critical subtasks (100Hz)
    //
    //###############################################################################
    // These tasks are run in a consequitive ripple to reduce the overhead on more
    // critical code. They are performed in reverse case order and will all complete
    // within the 10ms period.
    // Tasks are inhibitied by UsbState
    if (Task10ms > 0) {
      switch(Task10ms) {
        case 1: if (RotUpd) {MenuCheck();}; break;  // the encoder has been moved
        case 2: if (BUILD) {readSW0();} break;      // read rotary encoder switch SW0
      } Task10ms--; // count down the sub-tasks from 3 - 1 as we loop
    } else {
      if ((millis() - next10ms) >= 10) {
        next10ms = millis(); // reset the 10ms timer
        if (UsbActive > 0) {
          // one or more serial characters have been received
          // we should receive an ACK from the app every 500ms
          UsbActive--;  // reduce the 1 sec time-out
          if (keyVal != ACK) {ACK_Cnt = 0;} // remove transient glitches on serial link
          if ((!UsbState) && (ACK_Cnt == 3)) {
            // just received 3 ACK characters in non active state
            // so we enter a locked state, stopping anything that is running
            DispLock = true;  // trigger a lock-down
            UsbState = true;  // prevent re-triggering in the next cycle
            AnimSTOP = true;  // STOP either animation engines
            BootAnim = false; // clear boot up flag
          }
          if (UsbActive < 1) {
            // time-out reached, so free up display again
            UsbState = false; // drop the locked USB flag
            DispNOP = false;  // re-enable doFrameTasks()
          }
          Task10ms = 0; // inhibit tasks
        } else {Task10ms = 2;}  // restart tasks
      }// delay(1);       // return to RTOS, not yield() does not work here
    }
    // check RTOS loop timer
    if ((millis() - loopDelay) > 10) {
      // times out every 10ms for freeRTOS
      loopDelay = millis();
      vTaskDelay(1);
    }
  }
}

// ---------------------------------------------------------------------------------

void RunPOST() {
  // start-up initialisation processes  
  Serial.println("\n\n" + Release + "  " + Date + "\n");

  // initilaise RGB LED array
  FastLED.clear();
  FastLED.show(); // This sends the updated pixel color to the hardware.
  Serial.println("RGB LEDs cleared.");
  
  // Initialise display
  if (BUILD) {
    // build assumes display and rotary encoder are connected
    Display_Init();
    display.setBrightness(Brightness);
    Display_Intro();
    Serial.println("Display initialised");
    DispDel = 50;   // wait 2 seconds before displaying anything else
    DispClr = true; // clear the new screen as first action after delay
  } else {
    Serial.println("BUILD defines display as disconnected!");
  }

  // Initialise SD card reader
  // note ErrSD will return true if an error orrurs
  initSD();
  SDin = ErrSD; // retain the error status, mounted = false, not mounted = true

  // empty the serial buffer
  while (Serial.available() > 0) {keyVal = Serial.read();}
  
  // check for SW0 button being pressed when coming out of REST
  if (!digitalRead(sw0Pin)) {
    // SW0 is being pressed during RESET so force TEST mode
    TEST = true;
  }

  BootAnim = false;
  if (TEST) {
    // TEST mode
    Serial.println("TEST mode");
  } else {
    // normal mode
    Serial.println("Normal mode");
    MenuSet(0);
  }
  // run An0000.txt if it exists
  ClearAnimArray();
  readSDAnim("0000"); // loads /Anim/An0000.txt
  if (ErrFlg) {
    Serial.println("No animation script!");
  } else {
    setAnimMode(1); AnimPLAY = true; // start the animation engine
    Serial.println("Playing animation boot file from SD");
    BootAnim = true;  // flag running animation from boot up
    RepON = false;    // turn OFF any reporting when booting from SD
    Verbose = false;  // turn OFF verbose messaging
  }

  Serial.println("POST completed");
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values
  ACK_Cnt = 0;        // counter used in app link connection
  AnimBuild = false;  // == true if the animation engine is set to run
  AnimDel = 0;        // if > 0 then skip frame cycles
  AnimErr = 0;        // if > 0 it references an animation error
  AnimFrame = -1;     // animation frame loaded, -1 not loaded
  AnimIf = true;      // do next line, default = true
  AnimMode = -1;      // a pointer which determines which frame build process is used
  AnimPacket = false; // if == true then decode serial data as Anim packets
  AnimPAUSE = false;  // if == true then the current animation is paused
  AnimPLAY = false;   // == true if an animation has been set to play
  AnimPnt = 0;        // animation engines line pointer
  AnimSTEP = false;   // animation engine single STEP invokes a pause
  AnimSTOP = false;   // set == true to stop current animation
  AnimWait = 0;       // time delay in millis()
  BckGndFrame = -1;   // background frame loaded, -1 not loaded
  Border = false;     // if true then display has a border
  Bright = BrightDef; // default FastLED brightness level
  BrightAuto = true;  // if == true automatically increase brightness to max levels
  BrightCnt = 0;      // counter used in auto max
  BrightDef = 64;     // default FastLED brightness level
  Brightness = 192;   // OLED display brightness
  BrightNew = 255;    // latest FastLED brightness
  cmdMode = ' ';      // command mode
  cmdSgn = 1;         // cmdVal sign, 1 or -1
  cmdType = ' ';      // command mode type
  cmdVal = 0;         // value associated with a cmdType
  Col11 = 255;        // max colour
  Col34 = 192;        // 3/4 max colour
  Col12 = 128;        // 1/2 max colour
  Col14 = 64;         // 1/4 maxcolour
  DELcnt = 0;         // counts the delimeters in the Rx data stream
  DispBack = false;   // flag used to draw a back arrow on the display
  DispChng = false;   // flag invokes initialisation code
  DispClr = false;    // set == true to clear the memory contents
  DispCnt = 0;        // counter used in display updating
  DispDel = 0;        // >0 display delay counter, stops updates over I2C
  DispDisp = false;   // set = true to immediately display memory contents
  DispLock = false;   // set == true to lock the display
  DispMenu = 0;       // group type, 0 = sensors, 1 = tasks, etc
  DispMode = 0;       // the type of display for a given mode
  DispNOP = false;    // set == true to stop display I2C updates
  DispNow = 0;        // if > 0 display special modes
  DispPnt = 0;        // points at the current display mode, default = 0 batteries
  DispSub = 0;        // special display sub-task pointer
  DispTask = 0;       // special display task pointer
  Ext = ".txt";       // extension string. eg. '.txt'
  LED_Pnt = 0;        // pointer to LED to be lit
  LEDnoshow = false;  // temporary flag used to disable frame output
  loopDelay = 10;     // RTOS task timer
  Folder = "";        // folder name useed in filepath read/write
  ForGndFrame = -1;   // foreground frame loaded, -1 not loaded
  ForPnt = -1;        // For... Next loop stack pointer
  FrameBuild = false; // if == true then run a frame building task
  FrameCnt = 0;       // Main Task general counter/timer
  FrameDel = 0;       // Main task delay counter, if > 0 holds off tasks
  FrameMode = "";     // text describing the frame mode used by the editor
  FrameNext = 0;      // next MainTask to be performed after a delay
  FramePeriod = 40;   // frame time period in ms, default = 40 (25Hz)
  FramePeriodDef = 0; // default frame time period, if > 0
  FrameTask = 0;      // task pointer for frame construction
  FrameTitle = "";    // text describing the frame title used by the editor
  GfxP0 = NumLEDs - 1;// points to last pixel, top right corner
  GosubPnt = -1;      // Gosub stack pointer, empty = -1
  IcMax = 0;          // if > 0 the use IcMax/10 as current limit, default = 0
  IcOnce = false;     // set == true after each show for current measurement
  LED_Blink = false;  // flag toggled to create blinking effects
  LED_Cnt = 0;        // general counter used in LED tasks
  LED_Col = 0;        // general colour pointer used in LED tasks
  LED_Del = 0;        // LED Task delay, adds pauses into the processes
  LED_Inc = 1;        // general increment used in LED tasks
  LED_Max = 32;       // Max brightness level used by tasks
  LED_Nop = false;    // == true to block LED tasks, default = false
  LED_ON = false;     // built-in LED ON flag
  LEDshow = false;    // LED show flag for mouth LEDs
  LED_SubTask = 0;    // LED sub-task pointer
  LED_SwDwn = false;  // == true if a button switch has been pressed
  LED_Task = 0;       // LED task pointer
  LED_Val = 0;        // values used in LED tasks
  LevelCnt = 0;       // counts cumulative levels of all LEDs in output chain
  LevelOld = 0;       // previous cumulative levels of all LEDs in output chain
  Mem_Blu = 0;        // memory of current blue colour
  Mem_Grn = 0;        // memory of current green colour
  Mem_Red = 0;        // memory of current red colour
  MenuLevel = 0;      // pointer to menu level in the hierarchy, 0 = top, - N
  MenuPnt = 0;        // pointer to the current tier level, 0 - N
  MenuT0 = 0;         // menu choice range LL
  MenuT1 = 2;         // menu choice range LL
  Name = "";          // filename useed in filepath read/write, minus .txt extension
  PrintEn = true;     // if == false turn off verbose file system messaging
  REBOOT = false;     // set true when reboot is in progress to stop all tasks
  RepON = true;       // report ON/OFF flag, default = true
  ResetCnt = 0;       // timer used to invoke soft RESET
  RndCol = 255;       // previous random colour
  RotEncCnt = 32;     // rotary encoder counter
  RotUpd = true;      // == true when a count update has occured
  SchCnt = 0;         // animation scheduler array counter, default 0
  SchPnt = -1;        // animation scheduler array pointer, default -1
  ScrollTask = 0;     // == 0 to cause scroll initialisation
  sw0Cnt = 0;         // button switch counter
  sw0DwnTime = 0;     // button pressed down time
  sw0LastState = HIGH;// previous state of button switch, HIGH/LOW
  SW0_Nop = false;    // == true to block SW0 read tasks, default = false
  sw0State = HIGH;    // state of read button switch pin
  sw0Timer = 0;       // timer used to detemine button sequences
  Task10ms = 0;       // 10ms loop timer tasks
  Task40ms = 0;       // 40ms loop timer tasks
  TEST = TESTmode;    // normally false, set = true for assembly test mode
  Text = "";          // any text to be written to the display
  Text_SY = 0;        // Y position of scrolling text, default = 0
  Text_X = 0;         // text X co-ordinate
  Text_Y = 0;         // text Y co-ordinate
  Txt_Blu = 255;      // current text blue colour
  Txt_Grn = 0;        // current text green colour
  Txt_Red = 0;        // current text red colour
  UsbActive = 0;      // > 0 if serial data is received
  UsbState = false;   // == true if UsbActive
  Verbose = true;     // if == true send all text messages
  WaitUp = false;     // true if task is to wait for button release
  Word = "";          // any word extracted from Text to be written to the display
  WordRht = "";       // any remaining word extracted from Text to be written to the display
  WordSp = -1;        // pointer to the spaces in a text string between words

  SerialFlush();      // empty the serial input buffer
}

// ---------------------------------------------------------------------------------

void synchLoopTimers() {
  // reset main loop timers
  // a 2ms phase offset is introduced to prevent simultaneous triggering
  FrameTimer = millis() + 40; // variable frame timer offset by 18 initially
  next5ms = millis();         // 5ms loop timer not scewed
  next10ms = millis() + 1;    // 10ms loop timer scewed by 1ms
  next40ms = millis() + 2;    // 40ms loop timer scewed by 2ms
}

// --------------------------------------------------------------------------------
